#ifndef ENGINE_FRET_PI_H
#define ENGINE_FRET_PI_H

#include "fret_model_wrapper.h"
#include "../engine.h"
#include "../../types.h"
#include "../../storage/state_map.h"

#include <memory>
#include <vector>
#include <deque>

namespace fd {
namespace engines {
namespace fret {

template<typename Model, typename Reward, typename InnerEngine>
class FretPi : public Engine<Model, Reward>
{
public:
    FretPi(
        std::shared_ptr<Model> model,
        std::shared_ptr<Reward> reward,
        std::shared_ptr<InnerEngine> inner_engine);

    void prepare();
    ValueType get_reward(const State& state);
    Action get_policy_action(const State& state);

private:

    struct SCCInfo {
        static const unsigned UNDEFINED;

        SCCInfo() : index(UNDEFINED), onstack(0) {}
        SCCInfo(const SCCInfo&) = default;
        SCCInfo(SCCInfo&&) = default;
        SCCInfo& operator=(const SCCInfo&) = default;
        SCCInfo& operator=(SCCInfo&&) = default;

        unsigned index;
        unsigned lowlink : 31;
        unsigned onstack : 1;
    };

    struct ExpansionInfo {
        ExpansionInfo(const State& s, ValueType reward)
            : state(s)
            , reward(reward)
            , zero_reward_cycle(true)
            , permanent(true)
        {}

        State state;
        ValueType reward;
        bool zero_reward_cycle;
        bool permanent;
        std::vector<std::pair<ValueType, State> > successors;
    };

    std::shared_ptr<InnerEngine> m_inner_engine;

};


template<typename Model, typename Reward, typename InnerEngine>
FretPi<Model, Reward, InnerEngine>::FretPi(
    std::shared_ptr<Model> model,
    std::shared_ptr<Reward> reward,
    std::shared_ptr<InnerEngine> inner_engine)
    : Engine<Model, Reward>(model, reward)
    , m_inner_engine(inner_engine)
{}

template<typename Model, typename Reward, typename InnerEngine>
void
FretPi<Model, Reward, InnerEngine>::prepare()
{
    FD_LOG_MESSAGE("Starting trap elimination heuristic search wrapper...");
    std::deque<ExpansionInfo> policy_exploration;
    storage::StateMap<SCCInfo> scc_infos;
    std::deque<State> stack;
    State istate = this->m_transition_model->get_initial_state();
    bool eliminated_trap = true;
    unsigned fret_iteration = 0;
    while (eliminated_trap) {
        ++fret_iteration;
        eliminated_trap = false;
        m_inner_engine->reset_and_prepare();
        policy_exploration.emplace_back(
                istate,
                m_inner_engine->get_reward(istate));
        unsigned index = 0;
        ValueType succ_reward = 0;
        bool succ_permanent = true;
        bool succ_zero_cycle = true;
        while (!policy_exploration.empty()) {
            ExpansionInfo& expand = policy_exploration.back();
            SCCInfo& scc_info = scc_infos[expand.state];
            if (scc_info.index == SCCInfo::UNDEFINED) {
                scc_info.index = scc_info.lowlink = index++;
                Action policy = m_inner_engine->get_policy_action(expand.state);
                if (policy == UNDEFINED_ACTION) {
                    succ_reward = expand.reward;
                    policy_exploration.pop_back();
                    continue;
                }
                scc_info.onstack = 1;
                stack.push_front(expand.state);
                this->m_transition_model->get_successors(
                        expand.successors,
                        expand.state,
                        policy);
                assert(!expand.successors.empty());
            } else {
                const State& succ = expand.successors.back().second;
                SCCInfo& succ_info = scc_infos[succ];
                if (succ_info.onstack) {
                    scc_info.lowlink =
                        scc_info.lowlink < succ_info.lowlink ? scc_info.lowlink : succ_info.lowlink;
                    // TODO is this actually correct?
                    expand.zero_reward_cycle =
                        expand.zero_reward_cycle && succ_zero_cycle && value_equal(succ_reward, expand.reward);
                    expand.permanent = expand.permanent && succ_permanent;
                } else {
                    expand.permanent = false;
                }
            }
            while (!expand.successors.empty()) {
                const State& succ = expand.successors.back().second;
                SCCInfo& succ_info = scc_infos[succ];
                if (succ_info.index == SCCInfo::UNDEFINED) {
                    policy_exploration.emplace_back(succ, m_inner_engine->get_reward(succ));
                    break;
                } else if (succ_info.onstack) {
                    scc_info.lowlink =
                        scc_info.lowlink < succ_info.index ? scc_info.lowlink : succ_info.index;
                } else {
                    expand.permanent = false;
                }
                expand.successors.pop_back();
            }
            if (expand.successors.empty()) {
                if (scc_info.index == scc_info.lowlink) {
                    std::deque<State>::iterator end = stack.begin();
                    do {
                        scc_infos[*end].onstack = 0;
                    } while (*(end++) != expand.state);
                    if (expand.permanent && expand.zero_reward_cycle) {
                        // State new_state =
                            this->m_transition_model->add_trap(stack.begin(), end);
#if 0
                        std::cout << "trap [";
                        for (auto it = stack.begin(); it != end; it++) {
                            std::cout << (it != stack.begin() ? ", " : "")
                                 << *it;
                        }
                        std::cout << "] " << std::endl;
#endif
#if 0
                        if (new_state == 806) {
                            std::vector<Action> aops;
                            this->m_transition_model->get_applicable_actions(aops, new_state);
                            std::cout << " aops=[";
                            for (unsigned i = 0; i < aops.size(); i++) {
                                std::cout << (i > 0 ? ", " : "") << aops[i];
                            }
                            std::cout << "]" << std::endl;
                            for (unsigned i = 0; i < aops.size(); i++) {
                                std::cout << " successors[" << i << ", op=" << aops[i] << "] = [";
                                std::vector<std::pair<ValueType, State> > successors;
                                this->m_transition_model->get_successors(successors, new_state, aops[i]);
                                for (unsigned j = 0; j < successors.size(); j++) {
                                    std::cout << (j > 0 ? ", " : "") << successors[j].second;
                                }
                                std::cout << "]" << std::endl;
                            }
                        }
#endif
                        eliminated_trap = true;
                        // Action cur = 0;
                        // m_inner_engine->update_reward_get_policy(cur, new_state);
                    } else {
                        // for (auto it = stack.begin(); it != end; it++) {
                        //     Action old = m_inner_engine->get_policy_action(*it);
                        //     Action cur = old;
                        //     eliminated_trap = !m_inner_engine->update_reward_get_policy(cur, *it) || old != cur || eliminated_trap;
                        // }
                    }
                    stack.erase(stack.begin(), end);
                }
                succ_reward = expand.reward;
                succ_permanent = expand.permanent;
                succ_zero_cycle = expand.zero_reward_cycle;
                policy_exploration.pop_back();
            }
        }

        scc_infos.clear();

        FD_LOG_DEBUG_MESSAGE_T(
                "FRET iteration=" << fret_iteration << " reward=" << get_reward(istate));
    }
}

template<typename Model, typename Reward, typename InnerEngine>
ValueType
FretPi<Model, Reward, InnerEngine>::get_reward(const State& state)
{
    return m_inner_engine->get_reward(state);
}

template<typename Model, typename Reward, typename InnerEngine>
Action
FretPi<Model, Reward, InnerEngine>::get_policy_action(const State& state)
{
    return m_inner_engine->get_policy_action(state);
}


template<typename Model, typename Reward, typename InnerEngine>
const unsigned FretPi<Model, Reward, InnerEngine>::SCCInfo::UNDEFINED = -1;

}
}
}

#endif
