#include "types.h"
#include "cli/macros.h"
#include "utils/rng.h"
#include "utils/timer.h"

#include "sas_task/probabilistic_sas_task.h"
#include "models/transition_cache.h"
#include "models/sas/probabilistic_transition_model.h"
#include "models/sas/maxprob_reward_model.h"
#include "models/sas/minprob_reward_model.h"
#include "models/sas/expcost_reward_model.h"

#include "heuristics/heuristic.h"
#include "heuristics/dead_end_heuristic.h"
#include "heuristics/classic/hmax_dead_end_heuristic.h"

#include "engines/engine.h"

#include "engines/fret/fret_model_wrapper.h"
#include "engines/fret/fret_pi.h"
#include "engines/heuristic_depth_first_search.h"

#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <memory>
#include <iomanip>
#include <limits>
#include <algorithm>
#include <sstream>

#define FD_ENABLE_PROPERTY_MAXPROB 1
#define FD_ENABLE_PROPERTY_MINPROB 0
#define FD_ENABLE_PROPERTY_EXPCOST 0

namespace fd {

void
print_help_and_exit()
{
    std::cout << "Print this help message and exit:" << std::endl;
    std::cout << "  --help, -h" << std::endl << std::endl;

    std::cout << "Required parameters:" << std::endl;
    std::cout << "  --engine {fretpi-hdp}              Engine to use." << std::endl;
    std::cout << std::endl;

    std::cout << "Optional parameters:" << std::endl;
    bool sep = false;
    std::cout << "  --sas SAS                          Path to output.sas file. If not set, tries to read model from stdin." << std::endl;
    std::cout << "  --property {";
#if FD_ENABLE_PROPERTY_MAXPROB
    std::cout << "maxprob";
    sep = true;
#endif
#if FD_ENABLE_PROPERTY_MINPROB
    std::cout << (sep ? "," : "") << "minprob";
    sep = true;
#endif
#if FD_ENABLE_PROPERTY_EXPCOST
    std::cout << (sep ? "," : "") << "expcost,expsteps";
    sep = true;
#endif
    sep = !sep; // ignore not used
    std::cout << "}" << std::endl << "                                     Optimization property." << std::endl;

    std::cout << "  --heuristic {hmax}                 Dead-end detection heuristic to use." << std::endl;
    std::cout << "  --epsilon EPS                      Epsilon parameter." << std::endl;
    std::cout << "  --seed SEED                        Seed for random number generator." << std::endl;
    std::cout << "  --enable-caching                   Enable transition caching." << std::endl;
    std::cout << std::endl;

    FD_EXIT(cli::ExitCode::CLI_PARSE_ERROR);
}

void
assert_not_last(const int& i, const std::vector<std::string>& args) {
    if (i == 0) {
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, args[i] << " requires an argument");
    }
}

void
assert_empty(std::vector<std::string>& args)
{
    if (!args.empty()) {
        std::ostringstream out; out << "Unmatched arguments: \"";
        for (int i = args.size() - 1; i >= 0; i--) {
            out << (i + 1 < (int)args.size() ? " " : "") << args[i];
        }
        out << "\"";
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, out.str());
    }
}

void
pop(const int& i, std::vector<std::string>& args)
{
    args.erase(args.begin() + i);
}

void
pop2(const int& i, std::vector<std::string>& args)
{
    auto start = args.begin() + i;
    auto end = start + 2;
    args.erase(start, end);
}

template<typename Engine, typename Model, typename Reward>
int
solve(std::shared_ptr<Engine> engine,
      std::shared_ptr<Model> model,
      std::shared_ptr<Reward> reward)
{
    utils::Timer t;
    engine->prepare();
    FD_LOG_MESSAGE("Initial state reward: " 
        << std::setprecision(std::numeric_limits<long double>::digits10 + 1)
        << reward->transform_reward(
            engine->get_reward(
                model->get_initial_state())));
    model->print_statistics();
    FD_LOG_MESSAGE("Solver time: " << t);
    FD_LOG_MESSAGE("Total time: " << utils::g_timer);

    return 0;
}

template<typename Model, typename Reward>
int
main(
        std::vector<std::string>& args,
        std::shared_ptr<sas_task::ProbabilisticSASTask> task,
        std::shared_ptr<Model> model,
        std::shared_ptr<Reward> reward_model)
{

    std::shared_ptr<heuristics::Heuristic<Reward> > heuristic = nullptr;

    for (int i = args.size() - 1; i >= 0; i--) {
        if (args[i] == "--heuristic") {
            assert_not_last(i, args);
            if (args[i-1] != "hmax") {
                FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "No such heuristic " << args[i-1]);
            }
            std::shared_ptr<heuristics::classic::ClassicalPlanningHeuristic> hmax =
                std::make_shared<heuristics::classic::HmaxDeadEndHeuristic>(task->get_determinization());
            heuristic = std::make_shared<heuristics::DeadEndHeuristic<Reward> >(hmax, reward_model, &model->get_state_cache());
            --i;
            pop2(i, args);
            break;
        }
    }

    for (int i = args.size() - 1; i >= 0; i--) {
        if (args[i] == "--engine") {
            assert_not_last(i, args);
            if (args[i-1] == "fretpi-hdp") {
                FD_LOG_MESSAGE("Engine: FRET-pi with HDP");
                std::shared_ptr<engines::fret::FretModel<Model> > fret_model =
                    std::make_shared<engines::fret::FretModel<Model> >(model);
                std::shared_ptr<engines::hdfs::HeuristicDepthFirstSearch<engines::fret::FretModel<Model>, Reward> > hdp =
                    std::make_shared<engines::hdfs::HeuristicDepthFirstSearch<engines::fret::FretModel<Model>, Reward> >(
                            fret_model,
                            reward_model,
                            heuristic);
                std::shared_ptr<engines::fret::FretPi<engines::fret::FretModel<Model>, Reward, engines::hdfs::HeuristicDepthFirstSearch<engines::fret::FretModel<Model>, Reward> > > fret_pi =
                    std::make_shared<engines::fret::FretPi<engines::fret::FretModel<Model>, Reward, engines::hdfs::HeuristicDepthFirstSearch<engines::fret::FretModel<Model>, Reward> > >(
                            fret_model,
                            reward_model,
                            hdp);
                --i;
                pop2(i, args);
                assert_empty(args);
                return solve(
                        fret_pi,
                        model,
                        reward_model);
            } else {
                FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Unknown engine " << args[i-1]);
            }
            break;
        }
    }

    FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "No engine defined");
}

template<typename Reward>
int
main(
        std::vector<std::string>& args,
        std::shared_ptr<sas_task::ProbabilisticSASTask> task,
        std::shared_ptr<models::sas_model::ProbabilisticTransitionModel> model,
        bool cache_transitions)
{
    if (cache_transitions) {
        return main(
                args,
                task,
                std::make_shared<models::TransitionCacheModel<models::sas_model::ProbabilisticTransitionModel> >(model),
                std::make_shared<Reward>(task, model->get_state_cache()));
    } else {
        return main(args, task, model, std::make_shared<Reward>(task, model->get_state_cache()));
    }
}

template<typename Reward>
int
main(
        std::vector<std::string>& args,
        std::shared_ptr<sas_task::ProbabilisticSASTask> task,
        std::shared_ptr<models::sas_model::ProbabilisticTransitionModel> model,
        std::shared_ptr<Reward> rmodel,
        bool cache_transitions)
{
    if (cache_transitions) {
        return main(
                args,
                task,
                std::make_shared<models::TransitionCacheModel<models::sas_model::ProbabilisticTransitionModel> >(model),
                rmodel);
    } else {
        return main(args, task, model, rmodel);
    }
}

int
main(std::vector<std::string>& args)
{

    std::string sas_file;

    bool cache_transitions = false;
    for (int i = args.size() - 1; i >= 0; i--) {
        if (args[i] == "--epsilon") {
            assert_not_last(i, args);
            try {
                g_epsilon = std::stod(args[i-1]);
            } catch (const std::invalid_argument& ) {
                FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Unexpected argument " << args[i-1] << " after " << args[i]);
            } catch (const std::out_of_range& ) {
                FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Unexpected argument " << args[i-1] << " after " << args[i]);
            }
            --i;
            pop2(i, args);
        } else if (args[i] == "--seed") {
            assert_not_last(i, args);
            try {
                utils::g_rng.seed(std::stoi(args[i-1]));
            } catch (const std::invalid_argument& ) {
                FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Unexpected argument " << args[i-1] << " after " << args[i]);
            } catch (const std::out_of_range& ) {
                FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Unexpected argument " << args[i-1] << " after " << args[i]);
            }
            pop2(i, args);
        } else if (args[i] == "--enable-caching") {
            cache_transitions = true;
            pop(i, args);
        } else if (args[i] == "--sas") {
            assert_not_last(i, args);
            sas_file = args[i-1];
            --i;
            pop2(i, args);
        }
    }


    std::shared_ptr<sas_task::ProbabilisticSASTask> sas_task = nullptr;
    if (sas_file.size() > 0) {
        std::ifstream f(sas_file);
        if (f.is_open()) {
            sas_task = std::make_shared<sas_task::ProbabilisticSASTask>(f);
            f.close();
        } else {
            FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Could not open SAS file " << sas_file);
        }
    } else {
        sas_task = std::make_shared<sas_task::ProbabilisticSASTask>(std::cin);
    }

    if (sas_task == nullptr) {
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "No model given");
    }

    std::shared_ptr<models::sas_model::ProbabilisticTransitionModel> model =
        std::make_shared<models::sas_model::ProbabilisticTransitionModel>(sas_task);

    for (int i = args.size() - 1; i >= 0; i--) {
        if (args[i] == "--property") {
            assert_not_last(i, args);
#if FD_ENABLE_PROPERTY_MAXPROB
            FD_LOG_MESSAGE("Property: MaxProb");
            if (args[i-1] == "maxprob") {
                return main<models::sas_model::MaxProbRewardModel>(args, sas_task, model, cache_transitions);
            } 
#endif
#if FD_ENABLE_PROPERTY_MINPROB
            FD_LOG_MESSAGE("Property: MinProb");
            if (args[i-1] == "minprob") {
                return main<models::sas_model::MinProbRewardModel>(args, sas_task, model, cache_transitions);
            } 
#endif
#if FD_ENABLE_PROPERTY_EXPCOST
            if (args[i-1] == "expcost") {
                FD_LOG_MESSAGE("Property: ExpCost");
                return main<models::sas_model::ExpCostRewardModel>(
                        args,
                        sas_task,
                        model,
                        std::make_shared<models::sas_model::ExpCostRewardModel>(
                            models::sas_model::ExpCostRewardModel::DEFAULT,
                            inf,
                            sas_task,
                            model->get_state_cache()),
                        cache_transitions);
            } else if (args[i-1] == "expsteps") {
                FD_LOG_MESSAGE("Property: ExpSteps");
                return main<models::sas_model::ExpCostRewardModel>(
                        args,
                        sas_task,
                        model,
                        std::make_shared<models::sas_model::ExpCostRewardModel>(
                            models::sas_model::ExpCostRewardModel::ONE,
                            inf,
                            sas_task,
                            model->get_state_cache()),
                        cache_transitions);
            } 
#endif
            FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "Unknown property " << args[i-1]);
        }
    }

#if FD_ENABLE_PROPERTY_MAXPROB
    FD_LOG_MESSAGE("Property: MaxProb");
    return main<models::sas_model::MaxProbRewardModel>(args, sas_task, model, cache_transitions);
#else
    FD_LOG_ERROR_AND_EXIT(cli::ExitCode::CLI_PARSE_ERROR, "No property given");
#endif
}

}

int
main(int argc, const char** argv)
{
    std::vector<std::string> args;
    for (int i = 1; i < argc; i++) {
        args.emplace_back(argv[i]);
        if (args.back() == "-h" || args.back() == "--help") {
            fd::print_help_and_exit();
        }
    }
    std::reverse(args.begin(), args.end());
    return fd::main(args);
}
