#include "probabilistic_transition_model.h"
#include "../../storage/array_compressor.h"
#include "../../storage/compressed_array_hash_set.h"
#include "../../storage/conjunctive_condition_evaluator.h"
#include "../../sas_task/probabilistic_sas_task.h"
#include "../../sas_task/fact_pair.h"
#include "../../sas_task/action.h"
#include "../../sas_task/probabilistic_action.h"
#include "../../cli/macros.h"

#include <algorithm>
#include <iostream>

namespace fd {
namespace models {
namespace sas_model {

using storage::compression::CompressedArrayConstReference;
using storage::compression::CompressedArray;

ProbabilisticTransitionModel::ProbabilisticTransitionModel(
        const std::shared_ptr<sas_task::ProbabilisticSASTask> task)
    : m_task(task)
    , m_aops_generator(NULL)
    , m_state_cache(NULL)
{
    if (task->has_axioms()) {
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::UNSUPPORTED_FEATURE, "Axioms are currently not supported");
    }

    std::vector<int> domains(task->get_num_variables());
    for (unsigned var = 0; var < task->get_num_variables(); var++) {
        domains[var] = task->get_variable_domain_size(var);
    }
    std::vector<Action> refs(task->get_num_actions());
    std::vector<std::vector<std::pair<int, int> > > preconditions(task->get_num_actions());
    for (unsigned i = 0; i < task->get_num_actions(); i++) {
        refs[i] = i;
        const sas_task::ProbabilisticAction& a = task->get_action(i);
        preconditions[i].reserve(a.get_preconditions().size());
        for (unsigned j = 0; j < a.get_preconditions().size(); j++) {
            preconditions[i].emplace_back(a.get_preconditions()[j].var, a.get_preconditions()[j].value);
        }
    }

    FD_LOG_MESSAGE_T("constructing successor generator...");
    m_aops_generator = 
            new storage::ConjunctiveConditionEvaluator<Action>(std::move(
                    storage::ConjunctiveConditionEvaluator<Action>::create(
                        domains,
                        refs,
                        preconditions)));
    FD_LOG_MESSAGE_T("done constructing successor generator!");

    m_state_cache = 
            new storage::CompressedArrayHashSet(domains);

    FD_LOG_MESSAGE("Bytes per state: " << m_state_cache->get_num_bytes_per_array());


    m_stat_num_expansions = 0;
    m_stat_num_states_generated = 0;
}

ProbabilisticTransitionModel::~ProbabilisticTransitionModel()
{
    if (m_aops_generator != NULL) {
        delete(m_aops_generator);
    }
    if (m_state_cache != NULL) {
        delete(m_state_cache);
    }
}

State
ProbabilisticTransitionModel::get_initial_state()
{
    return m_state_cache->insert_get_id(
            m_task->get_initial_state_values());
}

void
ProbabilisticTransitionModel::get_applicable_actions(std::vector<Action>& aops, const State& state)
{
    m_stat_num_expansions++;
    static std::vector<int> uncompressed(m_task->get_num_variables());
    m_state_cache->get_uncompressed_array(uncompressed, state);
    m_aops_generator->evaluate(aops, uncompressed);

// #ifndef NDEBUG
//     std::cout << "Getting applicable operators for state#" << state << " = [";
//     for (unsigned var = 0; var < m_task->get_num_variables(); var++) {
//         std::cout << (var > 0 ? ", " : "")
//                   << m_task->get_fact_name(sas_task::FactPair(var, uncompressed[var]));
//     }
//     std::cout << "]" << std::endl;
//     std::cout << "    => aops = [";
//     for (unsigned i = 0; i < aops.size(); i++) {
//         std::cout << (i > 0 ? ", " : "") << m_task->get_action(aops[i]).get_name();
//     }
//     std::cout << "]" << std::endl;
// #endif
}

void
ProbabilisticTransitionModel::get_successors(
    std::vector<std::pair<ValueType, State> >& successors,
    const State& sid,
    const Action& aid)
{
    const sas_task::ProbabilisticAction& action = m_task->get_action(aid);
    const std::vector<sas_task::ProbabilisticOutcome>& outcomes = action.get_outcomes();
    successors.reserve(successors.size() + outcomes.size());
    CompressedArrayConstReference state =
        m_state_cache->get_compressed_array(sid);
    static CompressedArray succ(m_state_cache->get_compressor());

    m_stat_num_states_generated += outcomes.size();

// #ifndef NDEBUG
//         std::cout << "apply(" << m_task->get_action(aid).get_name() << ", state#" << sid << " = [";
//         for (unsigned var = 0; var < m_task->get_num_variables(); var++) {
//             std::cout << (var > 0 ? ", " : "")
//                     << m_task->get_fact_name(sas_task::FactPair(var, state[var]));
//         }
//         std::cout << "]):" << std::endl;
// #endif
    for (unsigned i = 0; i < outcomes.size(); i++) {
        const sas_task::ProbabilisticOutcome& out = outcomes[i];
        const std::vector<sas_task::Effect>& effs = out.get_effects();
        state.copy(succ.data());
        for (int j = effs.size() - 1; j >= 0; j--) {
            bool fires = true;
            const std::vector<sas_task::FactPair>& cond = effs[j].get_condition();
            for (int k = cond.size() - 1; k >= 0; k--) {
                if (state[cond[k].var] != cond[k].value) {
                    fires = false;
                    break;
                }
            }
            if (fires) {
                const sas_task::FactPair& e = effs[j].get_effect();
                succ.set(e.var, e.value);
            }
        }
        // TODO evaluate axioms
        State succ_id = m_state_cache->insert_compressed_get_id(succ.data());
        successors.emplace_back(out.get_probability(), succ_id);
// #ifndef NDEBUG
//         std::cout << "     -> state#" << succ_id << " = [";
//         for (unsigned var = 0; var < m_task->get_num_variables(); var++) {
//             std::cout << (var > 0 ? ", " : "")
//                     << m_task->get_fact_name(sas_task::FactPair(var, succ[var]));
//         }
//         std::cout << "]" << std::endl;
// #endif
    }

    // std::cout << "     -> outcomes = [";
    // ValueType sum = 0;
    // for (unsigned i = 0; i < successors.size(); i++) {
    //     sum += successors[i].first;
    //     std::cout << (i > 0 ? ", " : "") << "state" << successors[i].second;
    // }
    // std::cout << "]" << std::endl;
    // std::cout << "     -> psum = " << sum << std::endl;
    // assert(value_equal(sum, 1));
}

void
ProbabilisticTransitionModel::print_statistics() const
{
    FD_LOG_MESSAGE("Expansions: "   << m_stat_num_expansions);
    FD_LOG_MESSAGE("Registered states: " << m_state_cache->size());
    FD_LOG_MESSAGE("Generated states: "  << m_stat_num_states_generated);
}

const std::shared_ptr<sas_task::ProbabilisticSASTask>
ProbabilisticTransitionModel::get_underlying_task() const
{
    return m_task;
}

const storage::CompressedArrayHashSet&
ProbabilisticTransitionModel::get_state_cache() const
{
    return *m_state_cache;
}

void 
ProbabilisticTransitionModel::fill_probabilities(
        std::vector<std::pair<ValueType, State> >& successors,
        unsigned start,
        const State& ,
        const Action& a) const
{
    const sas_task::ProbabilisticAction& action = m_task->get_action(a);
    const std::vector<sas_task::ProbabilisticOutcome>& outcomes = action.get_outcomes();
    assert(successors.size() - start == action.get_outcomes().size());
    for (int i = outcomes.size() - 1; i >= (int)start; i--) {
        successors[i].first = outcomes[i - start].get_probability();
    }
}

unsigned
ProbabilisticTransitionModel::get_num_outcomes(
        const State& ,
        const Action& action) const
{
    return m_task->get_action(action).get_outcomes().size();
}

}
}
}


