#ifndef MODEL_TRANSITION_CACHE_H
#define MODEL_TRANSITION_CACHE_H

#include "../types.h"
#include "../storage/segmented_vector.h"
#include "../cli/macros.h"

#include <cstring>
#include <memory>
#include <vector>

namespace fd {
namespace storage {
class CompressedArrayHashSet;
}
namespace sas_task {
class ProbabilisticSASTask;
}

namespace models {

template<typename Model>
class TransitionCacheModel {
public:
    TransitionCacheModel(std::shared_ptr<Model> model);
    ~TransitionCacheModel() = default;

    State get_initial_state();

    void get_applicable_actions(std::vector<Action>& aops, const State& state);

    void get_successors(
            std::vector<std::pair<ValueType, State> >& successors,
            const State& state,
            const Action& action);

    unsigned get_num_cached_states() const;

    void print_statistics() const;

    const std::shared_ptr<sas_task::ProbabilisticSASTask> get_underlying_task() const;
    const storage::CompressedArrayHashSet& get_state_cache() const;

    std::shared_ptr<Model> get_underlying_model() const;

    void fill_probabilities(
            std::vector<std::pair<ValueType, State> >& successors,
            const State& state,
            const Action& action);
    unsigned get_num_outcomes(const State& state, const Action& action);
private:
    struct Cache {
        std::vector<Action> applicable;
        State** successors;
        Cache() : successors(NULL) {}
        ~Cache()
        {
            for (int i = applicable.size() - 1; i >= 0; i--) {
                delete[] successors[i];
            }
            delete[] successors;
        }
    };
    Cache& lookup_in_cache(const State& state);
    std::shared_ptr<Model> m_underlying_model;
    storage::SegmentedVector<Cache> m_cache;

    unsigned m_stat_num_cached_transitions;
    unsigned m_stat_num_aops_cache_hits;
    unsigned m_stat_num_succ_gen_cache_hits;
    std::size_t m_stat_approx_memory;
};

template<typename Model>
TransitionCacheModel<Model>::TransitionCacheModel(
        std::shared_ptr<Model> model)
    : m_underlying_model(model)
{
    m_stat_num_cached_transitions = 0;
    m_stat_num_aops_cache_hits = 0;
    m_stat_num_succ_gen_cache_hits = 0;
    m_stat_approx_memory = 0;
}

template<typename Model>
State
TransitionCacheModel<Model>::get_initial_state()
{
    return m_underlying_model->get_initial_state();
}

template<typename Model>
void
TransitionCacheModel<Model>::get_applicable_actions(
        std::vector<Action>& aops,
        const State& state)
{
    Cache& cached = lookup_in_cache(state);
    if (cached.successors == NULL) {
        m_underlying_model->get_applicable_actions(cached.applicable, state);
        cached.successors = new State*[cached.applicable.size()];
        memset(cached.successors, 0, cached.applicable.size() * sizeof(State*));
        m_stat_approx_memory += cached.applicable.size() * (sizeof(Action) + sizeof(State*));
    } else {
        m_stat_num_aops_cache_hits++;
    }
    aops.reserve(aops.size() + cached.applicable.size());
    for (Action i = 0; i < cached.applicable.size(); i++) {
        aops.push_back(i);
    }
}

template<typename Model>
void
TransitionCacheModel<Model>::get_successors(
        std::vector<std::pair<ValueType, State> >& successors,
        const State& state,
        const Action& idx)
{
    Cache& cached = lookup_in_cache(state);
    assert(idx < cached.applicable.size());
    assert(cached.successors != NULL);
    const Action& action = cached.applicable[idx];
    int start = successors.size();
    if (cached.successors[idx] == NULL) {
        m_underlying_model->get_successors(successors, state, action);
        State*& succs = cached.successors[idx];
        succs = new State[successors.size() - start];
        for (int i = successors.size() - 1; i >= start; i--) {
            succs[i - start] = successors[i].second;
        }
        m_stat_num_cached_transitions++;
        m_stat_approx_memory += (successors.size() - start) * sizeof(State*);
    } else {
        m_stat_num_succ_gen_cache_hits++;
        const State* succs = cached.successors[idx];
        unsigned size = m_underlying_model->get_num_outcomes(state, action);
        successors.reserve(start + size);
        for (unsigned i = 0; i < size; i++) {
            successors.emplace_back(0, succs[i]);
        }
        m_underlying_model->fill_probabilities(
                successors,
                start,
                state,
                action);

    }
}

template<typename Model>
void
TransitionCacheModel<Model>::print_statistics() const
{
    m_underlying_model->print_statistics();
    FD_LOG_MESSAGE("Cached transitions: " << m_stat_num_cached_transitions);
    FD_LOG_MESSAGE("Expansion cache hits: " << m_stat_num_aops_cache_hits);
    FD_LOG_MESSAGE("Transition cache hits: " << m_stat_num_succ_gen_cache_hits);
    FD_LOG_MESSAGE("Approximate transition cache memory usage: "
            << (((m_cache.size() * sizeof(Cache)) + m_stat_approx_memory) / 1024)
            << "KB");
}

template<typename Model>
const std::shared_ptr<sas_task::ProbabilisticSASTask>
TransitionCacheModel<Model>::get_underlying_task() const
{
    return m_underlying_model->get_underlying_task();
}

template<typename Model>
const storage::CompressedArrayHashSet&
TransitionCacheModel<Model>::get_state_cache() const
{
    return m_underlying_model->get_state_cache();
}

template<typename Model>
std::shared_ptr<Model>
TransitionCacheModel<Model>::get_underlying_model() const
{
    return m_underlying_model;
}

template<typename Model>
void
TransitionCacheModel<Model>::fill_probabilities(
        std::vector<std::pair<ValueType, State> >& successors,
        const State& state,
        const Action& idx)
{
    Cache& cached = lookup_in_cache(state);
    assert(idx < cached.applicable.size());
    return m_underlying_model->fill_probabilities(successors, state, cached.applicable[idx]);
}

template<typename Model>
unsigned
TransitionCacheModel<Model>::get_num_outcomes(
        const State& state,
        const Action& idx)
{
    Cache& cached = lookup_in_cache(state);
    assert(idx < cached.applicable.size());
    return m_underlying_model->get_num_outcomes(state, cached.applicable[idx]);
}

template<typename Model>
typename TransitionCacheModel<Model>::Cache&
TransitionCacheModel<Model>::lookup_in_cache(const State& state)
{
    if (state >= m_cache.size()) {
        m_cache.resize(state + 1);
    }
    return m_cache[state];
}

}
}

#endif
