#include "probabilistic_action.h"

#include <cassert>
#include <algorithm>
#include <sstream>
#include <iostream>
#include <map>

namespace fd {
namespace sas_task {

template<typename T>
bool my_less(const T& x, const T& y);

template<typename T>
bool my_equal(const T& x, const T& u);

template<typename T>
bool
my_less(const std::vector<T>& x, const std::vector<T>& y)
{
    if (x.size() < y.size()) {
        return true;
    } else if (x.size() > y.size()) {
        return false;
    }
    for (unsigned i = 0; i < x.size(); i++) {
        if (my_less(x[i], y[i])) {
            return true;
        } else if (my_less(y[i], x[i])) {
            return false;
        }
    }
    return false;
}

template<>
bool
my_less(const FactPair& p, const FactPair& q)
{
    return p.var < q.var || (p.var == q.var && p.value < q.value);
}

template<>
bool
my_less(const Effect& x, const Effect& y)
{
    return my_less(x.get_effect(), y.get_effect())
           || (!my_less(y.get_effect(), x.get_effect()) && my_less(x.get_condition(), y.get_condition()));
}

template<>
bool
my_less(const ProbabilisticOutcome& x,
        const ProbabilisticOutcome& y)
{
    // return x.cost < y.cost || (x.cost == y.cost && my_less(x.effects, y.effects));
    return my_less(x.get_effects(), y.get_effects());
}

template<typename T>
bool
my_equal(const std::vector<T>& x, const std::vector<T>& y)
{
    if (x.size() != y.size()) {
        return false;
    }
    for (int i = x.size() - 1; i >= 0; i--) {
        if (!my_equal(x[i], y[i])) {
            return false;
        }
    }
    return true;
}

template<>
bool
my_equal(const FactPair& p, const FactPair& q)
{
    return p.var == q.var && p.value == q.value;
}

template<>
bool
my_equal(const Effect& x, const Effect& y)
{
    return my_equal(y.get_effect(), x.get_effect()) && my_equal(x.get_condition(), y.get_condition());
}

template<>
bool
my_equal(const ProbabilisticOutcome& x,
         const ProbabilisticOutcome& y)
{
    // return x.cost == y.cost && my_equal(x.effects, y.effects);
    return my_equal(x.get_effects(), y.get_effects());
}

template<typename T>
struct Less
{
    bool
    operator()(const T& x, const T& y) const
    {
        return my_less(x, y);
    }
};

template<typename T>
struct Equal
{
    bool
    operator()(const T& x, const T& y) const
    {
        return my_equal(x, y);
    }
};

template<typename T>
struct PointerLess
{
    bool
    operator()(const T* x, const T* y) const
    {
        return my_less(*x, *y);
    }
};

Effect
canonical(const Effect& eff)
{
    std::vector<FactPair> conds(eff.get_condition());
    std::sort(conds.begin(), conds.end(), Less<FactPair>());
    conds.erase(std::unique(conds.begin(), conds.end(), Equal<FactPair>()),
                conds.end());
    return Effect(eff.get_effect(), std::move(conds));
}

ProbabilisticOutcome
ProbabilisticOutcome::canonical() const
{
    std::vector<Effect> effs(effects);
    std::sort(effs.begin(), effs.end(), Less<Effect>());
    effs.erase(std::unique(effs.begin(), effs.end(), Equal<Effect>()),
               effs.end());
    return ProbabilisticOutcome(std::move(effs), probability);
}

void
ProbabilisticOutcome::dump() const
{
    std::cout << "p=" << probability
              //<< " c=" << cost
              << " effs:" << std::endl;
    for (unsigned i = 0; i < effects.size(); i++) {
        std::cout << "    - [";
        for (unsigned j = 0; j < effects[i].get_condition().size(); j++) {
            std::cout
                << (j > 0 ? " & " : "")
                << effects[i].get_condition()[j].var
                << "=" << effects[i].get_condition()[j].value;
        }
        std::cout << "] -> " << effects[i].get_effect().var << "=" << effects[i].get_effect().value << std::endl;
    }
}

void
ProbabilisticAction::normalize()
{
    std::vector<ProbabilisticOutcome> canon;
    canon.reserve(outcomes.size());
    std::map<ProbabilisticOutcome*, ValueType, PointerLess<ProbabilisticOutcome> >
    o_to_p;
    for (unsigned i = 0; i < outcomes.size(); i++) {
        canon.emplace_back(outcomes[i].canonical());
        ProbabilisticOutcome* key = &canon.back();
        auto it = o_to_p.find(key);
        if (it != o_to_p.end()) {
            it->second += key->get_probability();
            canon.pop_back();
        } else {
            o_to_p[key] = key->get_probability();
        }
    }
    for (auto it = o_to_p.begin(); it != o_to_p.end(); it++) {
        it->first->set_probability(it->second);
    }
    outcomes.swap(canon);

#ifndef NDEBUG
    ValueType sum = 0;
    for (unsigned i = 0; i < outcomes.size(); i++) {
        sum += outcomes[i].get_probability();
    }
    assert(value_equal(sum, 1));
#endif
}


}
}

